package ru.dbotthepony.kommons.collect

import ru.dbotthepony.kommons.util.KOptional
import java.lang.ref.Reference
import java.util.*
import java.util.concurrent.CompletableFuture
import java.util.stream.Stream
import java.util.stream.StreamSupport

fun <T> Array<T>.stream(): Stream<T> = Arrays.stream(this)

@Suppress("unchecked_cast")
fun <T> Stream<T?>.filterNotNull(): Stream<T> {
	return filter { it != null } as Stream<T>
}

@Suppress("unchecked_cast")
inline fun <reified T> Stream<*>.filterIsInstance(): Stream<T> {
	return filter { it is T } as Stream<T>
}

inline fun <T> MutableIterable<Reference<out T>>.forValidRefs(fn: (T) -> Unit) {
	val iterator = iterator()

	for (value in iterator) {
		val get = value.get()

		if (get == null) {
			iterator.remove()
		} else {
			fn.invoke(get)
		}
	}
}

inline fun <T> MutableIterable<Reference<out T>>.forValidRefsBreak(fn: (T) -> Boolean) {
	val iterator = iterator()

	for (value in iterator) {
		val get = value.get()

		if (get == null) {
			iterator.remove()
		} else {
			if (fn.invoke(get)) {
				break
			}
		}
	}
}

fun <S, T> chainOptionalFutures(source: Iterable<S>, mapper: (S) -> CompletableFuture<KOptional<T>>): CompletableFuture<KOptional<T>> {
	val itr = source.iterator()

	if (itr.hasNext()) {
		var future = mapper.invoke(itr.next())

		for (v in itr)
			future = future.thenCompose { if (it.isPresent) CompletableFuture.completedFuture(it) else mapper.invoke(v) }

		return future
	}

	return CompletableFuture.completedFuture(KOptional.empty())
}

fun <T> Stream<T>.asIterable(): Iterable<T> {
	return object : Iterable<T> {
		override fun iterator(): Iterator<T> {
			return this@asIterable.iterator()
		}
	}
}

/**
 * Returns applicable index to put [element] into [List] determined by [comparator], optionally specifying ranges as [fromIndex] and [toIndex]
 *
 * If [List] is not sorted, result of this function is undefined
 */
fun <E> List<E>.searchInsertionIndex(element: E, comparator: Comparator<in E>, fromIndex: Int = 0, toIndex: Int = size): Int {
	require(toIndex >= fromIndex) { "Invalid range: to $toIndex >= from $fromIndex" }
	require(fromIndex >= 0) { "Invalid from index: $fromIndex" }
	require(toIndex >= 0) { "Invalid to index: $toIndex" }
	require(fromIndex <= size) { "Invalid from index: $fromIndex (list size $size)" }
	require(toIndex <= size) { "Invalid to index: $toIndex (list size $size)" }

	if (fromIndex == size || fromIndex == toIndex || comparator.compare(element, this[fromIndex]) <= 0) {
		return fromIndex
	}

	// линейный поиск если границы маленькие
	if (toIndex - fromIndex <= 10) {
		for (i in fromIndex + 1 until toIndex) {
			val compare = comparator.compare(element, this[i])

			if (compare <= 0) {
				return i
			}
		}

		return size
	} else {
		// двоичный поиск
		var lower = fromIndex
		var upper = toIndex - 1

		while (upper - lower >= 10) {
			val middle = (upper + lower) / 2
			val compare = comparator.compare(element, this[middle])

			if (compare == 0) {
				return middle
			} else if (compare < 0) {
				upper = middle
			} else {
				lower = middle
			}
		}

		return searchInsertionIndex(element, comparator, lower, upper + 1)
	}
}

/**
 * Inserts [element] into [MutableList] at index determined by [comparator]
 *
 * If [MutableList] is not sorted, result of this function is undefined
 */
fun <E> MutableList<E>.addSorted(element: E, comparator: Comparator<in E>) {
	add(searchInsertionIndex(element, comparator), element)
}

private object NaturalComparator : Comparator<Comparable<Any>> {
	override fun compare(o1: Comparable<Any>, o2: Comparable<Any>): Int {
		return o1.compareTo(o2)
	}
}

/**
 * Inserts [element] into [MutableList] at index determined by comparing values themselves
 *
 * If [MutableList] is not sorted, result of this function is undefined
 */
fun <E : Comparable<E>> MutableList<E>.addSorted(element: E) {
	add(searchInsertionIndex(element, NaturalComparator as Comparator<in E>), element)
}

fun <E> MutableCollection<E>.addAll(elements: Iterator<E>) {
	for (item in elements) {
		add(item)
	}
}

fun <E> MutableCollection<E>.addAll(elements: Stream<out E>) {
	for (item in elements) {
		add(item)
	}
}

fun <E> Iterable<E>.stream(): Stream<out E> {
	return StreamSupport.stream(this.spliterator(), false)
}

fun <E> Iterator<E>.stream(): Stream<out E> {
	return StreamSupport.stream(Spliterators.spliteratorUnknownSize(this, 0), false)
}
